/*
 * Copyright (c) 2017 Gregory Penn.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#ifndef GDAL_AFFINE_POINT_TRANSFORMATION_HPP
#define GDAL_AFFINE_POINT_TRANSFORMATION_HPP

#include <array>
#include <gdal.h>
#include <gdal/point.hpp>
#include <gdal/geographic_srs.hpp>
#include <gdal/grid_srs.hpp>
#include <eigen3/Eigen/Dense>
#include <eigen3/Eigen/QR>


namespace gdal
{
  template < typename SourceSRS >
  class affine_transform;

  template <>
  class affine_transform<geographic_srs>
  {
    using grid_coords = Eigen::Matrix<int, 2, 1>;
    using geog_coords = Eigen::Matrix<double, 2, 1>;
    using matrix_type = Eigen::Matrix<double, 2, 2>;
    using qr_type = Eigen::ColPivHouseholderQR<matrix_type>;

  public:
    affine_transform () = default;

    template < typename Coeffs >
    affine_transform (geographic_srs srs, grid_srs grid, Coeffs coeffs)
      : srs_(srs)
      , grid_(grid)
      {
        origin_ << coeffs[0], coeffs[3];
        matrix_type matrix;
        matrix << coeffs[1], coeffs[2], coeffs[4], coeffs[5];
        qr_type qr(matrix);
        matrix_ = qr;
      }

    auto operator () (point<geographic_srs> pt)
      {
        assert(pt.srs() == srs());

        geog_coords coords;
        coords << pt.x(), pt.y();

        auto transformed_coords = matrix_.solve(coords - origin_);
        auto floored_coords = transformed_coords.array().floor();

        auto grid_coords = floored_coords.cast<int>().eval();
        return point<grid_srs>(grid_coords, grid_);
      }

  public:
    auto origin () const { return origin_; }
    auto matrix () const { return matrix_; }
    grid_srs grid () const { return grid_; }
    geographic_srs srs () const { return srs_; }

  private:
    geog_coords origin_;
    qr_type matrix_;
    geographic_srs srs_;
    grid_srs grid_;

  };

  template <>
  class affine_transform<grid_srs>
  {
    using grid_coords = Eigen::Matrix<int, 2, 1>;
    using geog_coords = Eigen::Matrix<double, 2, 1>;
    using matrix_type = Eigen::Matrix<double, 2, 2>;

  public:
    affine_transform () = default;

    template < typename Coeffs >
    affine_transform (geographic_srs srs, grid_srs grid, Coeffs coeffs)
      : srs_(srs)
      , grid_(grid)
      {
        origin_ << coeffs[0], coeffs[3];
        matrix_ << coeffs[1], coeffs[2], coeffs[4], coeffs[5];
      }

  public:
    point<geographic_srs> operator () (point<grid_srs> pt)
      {
        grid_coords coords;
        coords << pt.x(), pt.y();

        auto transformed_coords = (matrix_ * coords.cast<double>() + origin_).eval();

        return point<geographic_srs>(transformed_coords[0]
                                     , transformed_coords[1]
                                     , srs_);

      }

  public:
    auto origin () const { return origin_; }
    auto matrix () const { return matrix_; }
    auto srs    () const { return srs_; }
    auto grid   () const { return grid_; }

  private:
    geog_coords    origin_;
    matrix_type    matrix_;
    geographic_srs srs_;
    grid_srs       grid_;

  };

} // namespace gdal
#endif
