/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file
/// \brief Construct a strategy for a uniform distribution
///

#ifndef GRAPH_MODEL_DISTRIBUTION_UNIFORM_UNIFORM_DISTRIBUTION_HPP
#define GRAPH_MODEL_DISTRIBUTION_UNIFORM_UNIFORM_DISTRIBUTION_HPP 1

#include <boost/hana/map.hpp>
#include <graph_model/distribution/uniform/uniform_strategy.hpp>
#include <graph_model/distribution/parameter/make_parameter_default.hpp>
#include <graph_model/distribution/parameter/named_parameter.hpp>
#include <graph_model/distribution/parameter/parameter_pack.hpp>
#include <graph_model/distribution/parameter/lower_bound.hpp>
#include <graph_model/distribution/parameter/upper_bound.hpp>
#include <graph_model/distribution/parameter/variate.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace uniform
    {

      namespace detail
      {

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 parameter::named_parameter<Key,Value>,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged>,
	 Value&&,
	 Parameters&& ...);

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack)
	{ return make_uniform_strategy(pack); }

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Key, typename Value
		   , typename ... Parameters
		   >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 parameter::named_parameter<Key,Value> parameter,
	 Parameters&& ... parameters)
	{
	  return uniform_distribution
	    (pack.add(parameter),
	     std::forward<Parameters>(parameters)...);
	}

	template < typename Tags, typename Defaults, typename Tagged, typename Untagged
		   , typename Value
		   , typename ... Parameters
		   >
	auto uniform_distribution
	(parameter::parameter_pack<Tags,Defaults,Tagged,Untagged> pack,
	 Value&& value,
	 Parameters&& ... parameters)
	{
	  return uniform_distribution
	    (pack.add(std::forward<Value>(value)),
	     std::forward<Parameters>(parameters)...);
	}

      }	// namespace detail

      template < typename ... Parameters >
      auto uniform_distribution (Parameters&& ... parameters)
      {
	using boost::hana::make_map;
	using parameter::make_parameter;
	using parameter::make_parameter_pack;
	return detail::uniform_distribution
	  (make_parameter_pack
	   (make_parameter<parameter::tag::lower_bound>().default_value(0),
	    make_parameter<parameter::tag::upper_bound>().default_value(1),
	    make_parameter<parameter::tag::variate>().default_value(0.5)),
	   std::forward<Parameters>(parameters)...);
      }

    } // namespace uniform

  } // namespace distribution

} // namespace graph_model

#endif
