/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_STRATEGY_OWNER_BASE_HPP
#define GRAPH_MODEL_VERTEX_STRATEGY_OWNER_BASE_HPP 1

#include <cassert>
#include <memory>
#include <type_traits>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Manage a strategy object
    /// \tparam Derived The type of derived class
    /// \tparam StrategyFactory The type of factory for generating strategy objects
    ///
    /// The behavior of many of the vertex classes is determined by a
    /// strategy that may be substituted at runtime.  This class
    /// manages the strategy object for a vertex class.  The \c
    /// reset_strategy() member function substitutes (or removes
    /// altogether) the strategy object.  The \c strategy() member
    /// function returns a reference to the strategy object for the
    /// convenience of other base classes using CRTP.
    ///
    /// The \tparam StrategyFactory class models the \c
    /// StrategyFactoryConcept (yet to be defined), which implements
    /// two static \c apply() member functions for constructing
    /// strategy objects either from nothing or from a policy object.
    ///
    /// \bug Define the \c StrategyFactoryConcept class.
    ///
    template < typename Derived, typename StrategyFactory >
    class strategy_owner_base
    {
      using self = strategy_owner_base;
      using pointer = typename StrategyFactory::pointer;

    public:
      strategy_owner_base () = default;
      
      template < typename Policy >
      explicit strategy_owner_base (Policy&& policy)
	: p_(initialize(std::forward<Policy>(policy)))
      {}

      template < typename Policy >
      auto& operator = (Policy&& policy)
      {
	p_ = initialize(std::forward<Policy>(policy));
	return derived();
      }

      auto& reset_strategy () { p_.reset(); return derived(); }

      template < typename Policy >
      auto& reset_strategy (Policy&& policy)
      {
	p_ = initialize(std::forward<Policy>(policy));
	return derived();
      }

    public:			// XXX - for CRTP access only
      auto& strategy () { assert(p_); return *p_; }
      auto& strategy () const { assert(p_); return *p_; }

    private:
      static auto initialize () { return StrategyFactory::apply(); }

      template < typename Policy >
      static auto initialize (Policy&& p)
      { return StrategyFactory::apply(std::forward<Policy>(p)); }

    private:
      Derived& derived () { return static_cast<Derived&>(*this); }
      Derived const& derived () const { return static_cast<Derived const&>(*this); }

    private:
      pointer p_ = initialize();
    };

  } // namespace vertex

} // namespace graph_model

#endif
