/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef IDENTIFIERS_IDENTIFIER_HPP
#define IDENTIFIERS_IDENTIFIER_HPP 1

#include <identifiers/identifier_fwd.hpp>
#include <iosfwd>
#include <string>
#include <boost/lexical_cast.hpp>
#include <boost/operators.hpp>
#include <mpl/enable_constructor.hpp>

namespace identifiers
{
  
  template < typename Identifier
	     , typename Domain
	     >
  class identifier
    : boost::totally_ordered<identifier<Identifier,Domain>
			     , boost::totally_ordered<identifier<Identifier,Domain>,Identifier
						      > >
  {
    using self = identifier;

  public:
    template < typename ... Args
	       , typename = mpl::enable_constructor_t<self,Args...>
	       >
    identifier (Args&& ... args)
      : id_(initialize(std::forward<Args>(args)...))
    {}

    explicit operator bool () const { return empty(); }

    bool empty () const { return id_.empty(); }

    std::ostream& print (std::ostream& os) const { return os << id_; }

    friend bool operator == (self const& lhs, self const& rhs)
    { return lhs.id_ == rhs.id_; }

    friend bool operator < (self const& lhs, self const& rhs)
    { return lhs.id_ < rhs.id_; }

    friend bool operator == (self const& lhs, Identifier rhs)
    { return lhs.id_ == rhs; }

    friend bool operator < (self const& lhs, Identifier rhs)
    { return lhs.id_ < rhs; }

  private:
    template < typename ... Args >
    static Identifier initialize (Args&& ... args)
    { return Identifier(std::forward<Args>(args)...); }

  private:
    Identifier id_;
  };

  template < typename Domain >
  class identifier< std::string,Domain >
    : boost::totally_ordered<identifier<std::string,Domain>
			     , boost::totally_ordered<identifier<std::string,Domain>,std::string
						      > >
  {
    using self = identifier;

  public:
    template < typename ... Args
	       , typename = mpl::enable_constructor_t<self,Args...>
	       >
    identifier (Args&& ... args)
      : id_(initialize(std::forward<Args>(args)...))
    {}

    explicit operator bool () const { return empty(); }

    bool empty () const { return id_.empty(); }

    std::ostream& print (std::ostream& os) const { return os << id_; }

    friend bool operator == (self const& lhs, self const& rhs)
    { return lhs.id_ == rhs.id_; }

    friend bool operator < (self const& lhs, self const& rhs)
    { return lhs.id_ < rhs.id_; }

    friend bool operator == (self const& lhs, std::string rhs)
    { return lhs.id_ == rhs; }

    friend bool operator < (self const& lhs, std::string rhs)
    { return lhs.id_ < rhs; }

  private:
    static std::string initialize () { return std::string(); }

    template < typename T, typename ... U >
    static std::string initialize (T&& t, U&& ... u)
    {
      return boost::lexical_cast<std::string>(std::forward<T>(t))
	+ initialize(std::forward<U>(u)...);
    }

  private:
    std::string id_;
  };

  template < typename Domain >
  class identifier< void,Domain >
    : public std::string
  // , boost::totally_ordered<identifier<void,Domain>
  // , boost::totally_ordered<identifier<void,Domain>,std::string
  //  > >
  {
    using self = identifier;
    using base = std::string;

  public:
    template < typename ... Args
	       , typename = mpl::enable_constructor_t<self,Args...>
	       >
    identifier (Args&& ... args)
      : base(initialize(std::forward<Args>(args)...))
    {}

    explicit operator bool () const { return this->empty(); }

    std::ostream& print (std::ostream& os) const { return os << static_cast<base>(*this); }

  private:
    static std::string initialize () { return std::string(); }

    template < typename T, typename ... U >
    static std::string initialize (T&& t, U&& ... u)
    {
      return boost::lexical_cast<std::string>(std::forward<T>(t))
	+ initialize(std::forward<U>(u)...);
    }
  };

  template < typename Identifier, typename Domain >
  std::ostream&
  operator << (std::ostream& os, identifier<Identifier,Domain> const& x)
  { return x.print(os); }

  template < typename ... Args >
  auto make_identifier (Args&& ... args)
  { return identifier<>(std::forward<Args>(args)...); }

  template < typename Identifier, typename ... Args >
  auto make_identifier (Args&& ... args)
  {
    using domain_type = traits::domain_t<Identifier>;
    return identifier<Identifier,domain_type>(std::forward<Args>(args)...);
  }

} // namespace identifiers

#endif
