/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE dataframe_factor

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

#include <dataframe/factor.hpp>
#include <set>
#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<int,long,float,double>;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,T,test_types)
{
  using factor_type = dataframe::factor<T>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_construction,T,test_types)
{
  using factor_type = dataframe::factor<T>;
  factor_type factor;
  BOOST_CHECK(factor.empty());
  BOOST_CHECK_EQUAL(factor.size(),0);
  BOOST_CHECK(factor.begin() == factor.end());
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_value,T,test_types)
{
  using factor_type = dataframe::factor<T>;

  factor_type factor;
  factor.insert(42);

  BOOST_CHECK(!factor.empty());
  BOOST_CHECK_EQUAL(factor.size(),1);
  BOOST_CHECK(factor.begin() != factor.end());

  BOOST_CHECK_EQUAL(*factor.begin(),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_range,T,test_types)
{
  using factor_type = dataframe::factor<T>;
  using vector_type = std::vector<T>;

  vector_type vector = { 0, 2, 1 };
  BOOST_CHECK_EQUAL(vector.size(),3);

  factor_type factor;
  factor.insert(vector.begin(),vector.end());

  BOOST_CHECK(!factor.empty());
  BOOST_CHECK_EQUAL(factor.size(),vector.size());
  BOOST_CHECK(factor.begin() != factor.end());

  BOOST_CHECK_EQUAL(*factor.begin(),0);
  BOOST_CHECK_EQUAL(*++factor.begin(),1);
  BOOST_CHECK_EQUAL(*++++factor.begin(),2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_make_factor,T,test_types)
{
  using vector_type = std::vector<T>;

  vector_type vector = { 0, 2, 1 };
  BOOST_CHECK_EQUAL(vector.size(),3);

  auto factor = dataframe::make_factor(vector,2);

  BOOST_CHECK(!factor.empty());
  BOOST_CHECK_EQUAL(factor.size(),vector.size());
  BOOST_CHECK(factor.begin() != factor.end());

  BOOST_CHECK_EQUAL(*factor.begin(),0);
  BOOST_CHECK_EQUAL(*++factor.begin(),1);
  BOOST_CHECK_EQUAL(*++++factor.begin(),2);

  BOOST_CHECK_EQUAL(factor.index(),2);
  BOOST_CHECK_EQUAL(factor.value(),2);
  BOOST_CHECK_EQUAL(static_cast<T>(factor),2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_equality,T,test_types)
{
  using vector_type = std::vector<T>;

  vector_type vector = { 0, 2, 1 };
  BOOST_CHECK_EQUAL(vector.size(),3);

  auto factor_0 = dataframe::make_factor(vector,0);
  auto factor_1 = dataframe::make_factor(vector,1);

  BOOST_CHECK(factor_0 == factor_0);
  BOOST_CHECK(factor_0 != factor_1);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_less_than,T,test_types)
{
  using vector_type = std::vector<T>;

  vector_type vector = { 0, 2, 1 };
  BOOST_CHECK_EQUAL(vector.size(),3);

  auto factor_0 = dataframe::make_factor(vector,0);
  auto factor_1 = dataframe::make_factor(vector,1);

  BOOST_CHECK(factor_0 < factor_1);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_conversion,T,test_types)
{
  using factor_type = dataframe::factor<T>;
  using other_factor_type = dataframe::factor<T,std::less<T>,std::set<T>>;
  using vector_type = std::vector<T>;

  vector_type vector = { 0, 2, 1 };
  BOOST_CHECK_EQUAL(vector.size(),3);

  other_factor_type other;
  other.insert(vector.begin(),vector.end());
  other.value(2);

  factor_type factor (other);

  BOOST_CHECK(!factor.empty());
  BOOST_CHECK_EQUAL(factor.size(),other.size());
  BOOST_CHECK_EQUAL(*factor.begin(),*other.begin());
  BOOST_CHECK_EQUAL(factor.index(),other.index());

  BOOST_CHECK(factor == other);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_merge,T,test_types)
{
  using factor_type = dataframe::factor<T>;
  using vector_type = std::vector<T>;

  auto factor_0 = dataframe::make_factor(std::vector<T>({0,1,2}),1);
  auto factor_1 = dataframe::make_factor(std::vector<T>({3,4,5}),4);
  
  BOOST_CHECK_EQUAL(factor_0,1);
  BOOST_CHECK_EQUAL(factor_1,4);
  BOOST_CHECK(factor_0.begin() != factor_1.begin());
  BOOST_CHECK(factor_0.end() != factor_1.end());

  merge(factor_0,factor_1);

  BOOST_CHECK_EQUAL(factor_0,1);
  BOOST_CHECK_EQUAL(factor_1,4);
  BOOST_CHECK(factor_0.begin() == factor_1.begin());
  BOOST_CHECK(factor_0.end() == factor_1.end());
}
