/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_examples_normal_sample_mean

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <iostream>
#include <map>

#include <boost/filesystem/fstream.hpp>
#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>

#include <dataframe/dataframe.hpp>
#include <dataframe/element.hpp>
#include <dataframe/read_dataframe.hpp>

#include <graph_model/MCMC/mcmc_signal_policy.hpp>
#include <graph_model/MCMC/metropolis_sampler.hpp>
#include <graph_model/distribution/bernoulli/bernoulli_distribution.hpp>
#include <graph_model/distribution/normal/normal_distribution.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/factor.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/vertex/markov_blanket.hpp>
#include <graph_model/vertex/sample.hpp>
#include <graph_model/vertex/summarized_random_variable.hpp>

#include <identifiers/x.hpp>
#include <mcmc/mcmc_tally.hpp>

#include <logging/initialize.hpp>

#include <boost/mpl/pair.hpp>
#include <boost/mpl/vector.hpp>

// Use a lognumeric type to represent probabilities
//
#if USE_LOGNUMERIC_PROBABILITY_TYPE
#include <sample_space/numeric/lognumeric.hpp>
#endif	// USE_LOGNUMERIC_PROBABILITY_TYPE

using test_types = boost::mpl::vector<
  boost::mpl::pair<double,long double>
  >;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_graph_model,Pair,test_types)
{
  // define some types
  using variate_type = typename Pair::first;
  using probability_type = typename Pair::second;

  using ifstream_type = boost::filesystem::ifstream;
  using ofstream_type = boost::filesystem::ofstream;

  using factor_type = graph_model::vertex::factor<probability_type>;
  using factor_map_type = std::map<int,factor_type>;

  using expression_type = graph_model::vertex::expression<variate_type>;
  using data_map_type = std::map<int,expression_type>;

  using severity_level = boost::log::trivial::severity_level;
  using logger_type = boost::log::sources::severity_logger<severity_level>;

  // simplify references to namespace-qualified identifiers
  using graph_model::distribution::bernoulli::bernoulli_distribution;
  using graph_model::distribution::normal::normal_distribution;
  using graph_model::vertex::lit;
  using graph_model::vertex::make_summarized_random_variable;

  namespace accumulators = boost::accumulators;
  namespace tag = graph_model::signals::tag;
  namespace parameter = graph_model::distribution::parameter;
  namespace MCMC = graph_model::MCMC;

  logging::initialize();
  logger_type debug_log;

  int MCMC_sample_size = MCMC_SAMPLE_SIZE;

  // construct the random variable to estimate
  auto p = make_summarized_random_variable<variate_type,probability_type>();
  p.reset_identifier("p");
#if defined(INITIAL_P)
  p.reset_variate(INITIAL_P);
#endif	// defined(INITIAL_P)

  // construct a dataframe of observations
  ifstream_type is("samples.dat");
  auto dataframe = dataframe::make_dataframe();
  is >> dataframe;

  // construct random variables for the observations and corresponding factors
  data_map_type data;
  factor_map_type factor;
  for (auto row : dataframe)
    {
      data[row]
	.reset_identifier(identifiers::x(row))
	.reset_strategy(dataframe::make_element<variate_type>(dataframe,row,"samples"))
	.add_vertex(factor[row]);
      factor[row]
	.reset_identifier("factor ",row)
	.reset_strategy(bernoulli_distribution(parameter::variate=data[row],
					       parameter::probability=p))
	.markov_blanket(p);
    }

  // construct the MCMC components
  auto mcmc_tally = mcmc::make_mcmc_tally();
  auto p_sampler
    = metropolis_sampler
    (markov_blanket(p),
     MCMC::make_mcmc_signal_policy()
     .template connect_signal<tag::iteration>
     ([]{ static std::size_t i = 0; if (i == 0) std::cout << "iteration: "; if (i % 100 == 0) std::cout << " " << i << std::flush; ++i; })
     .template connect_signal<tag::accept>(mcmc_tally.make_acceptor())
     .template connect_signal<tag::reject>(mcmc_tally.make_rejector()))
    .add_rv(p,lit(p) + sample(normal_distribution(0,0.1)));
#if !defined(INITIAL_P)
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "p_sampler.initialize()";
  p_sampler.initialize();
  p_sampler.commit();
#endif	// !defined(INITIAL_P)

  // construct an output stream for the MCMC samples and generate samples
  ofstream_type ofstream("mcmc.dat");
  ofstream << "i\tMCMC_sample\tmean" << std::endl;
  ofstream << 0 << "\t" << p.variate() << "\t" << accumulators::mean(p.accumulator()) << std::endl;
  GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC samples:";
  for (int i = 0; i < MCMC_sample_size; ++i)
    {
      GLOBAL_LOG(debug_log,boost::log::trivial::trace) << "MCMC sample: " << i;
      p_sampler.update();
      p_sampler.commit();
      ofstream << i+1 << "\t" << p.variate() << "\t" << accumulators::mean(p.accumulator()) << std::endl;
    }
  std::cout << std::endl;

  // report summary statistics
  std::cout << "estimated p:   " << accumulators::mean(p.accumulator()) << std::endl;
  std::cout << "estimated mean: " << accumulators::mean(p.accumulator()) << std::endl;
  std::cout << "MCMC tally: " << mcmc_tally << std::endl;
  std::cout << std::endl;
}
