/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_utility_function_collection

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/utility/function_collection.hpp>
#include <boost/operators.hpp>

struct function
  : boost::less_than_comparable<function>
{
  function (int i)
    : i_{i}
  {}

  int operator () () const { return i_; }

private:
  int i_;
};

bool operator < (function left, function right) { return left() < right(); }
auto& operator << (std::ostream& os, function f) { return os << f(); }

struct copyable
  : boost::less_than_comparable<copyable>
{
  copyable (int i) : i_(i) {}
  copyable (copyable const&) = default;
  copyable& operator = (copyable const&) = default;

  int operator () () const { return i_; }

  friend auto operator < (copyable const& left, copyable const& right)
  { return left.i_ < right.i_; }

  friend std::ostream& operator << (std::ostream& os, copyable const& m)
  { return os << m.i_; }

private:
  int i_;
};

struct moveable
  : boost::less_than_comparable<moveable>
{
  moveable (int i) : i_(i) {}
  moveable (moveable&&) = default;
  moveable& operator = (moveable&&) = default;

  int operator () () const { return i_; }

  friend auto operator < (moveable const& left, moveable const& right)
  { return left.i_ < right.i_; }

  friend std::ostream& operator << (std::ostream& os, moveable const& m)
  { return os << m.i_; }

private:
  int i_;
};

#include <boost/mpl/vector.hpp>

using copyable_test_types = boost::mpl::vector<function,copyable>;
using moveable_test_types = boost::mpl::vector<function,copyable,moveable>;

BOOST_AUTO_TEST_CASE(test_default_constructor)
{
  using signature = int();
  using function_collection_type = graph_model::utility::function_collection<signature>;
  function_collection_type f;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_lvalue,T,copyable_test_types)
{
  using signature = int();
  using function_collection_type = graph_model::utility::function_collection<signature>;

  T zero {0};
  T one {1};
  T three {3};

  function_collection_type f;
  f.insert(zero);
  f.insert(three);
  f.insert(one);
  f.insert(zero);
  f.insert(zero);

  auto i = f.begin<T>();
  BOOST_CHECK_EQUAL((*i++)(),0);
  BOOST_CHECK_EQUAL((*i++)(),3);
  BOOST_CHECK_EQUAL((*i++)(),1);
  BOOST_CHECK_EQUAL((*i++)(),0);
  BOOST_CHECK_EQUAL((*i++)(),0);
  BOOST_CHECK(i == f.end<T>());
}

#if XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED
BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_rvalue,T,moveable_test_types)
{
  using signature = int();
  using function_collection_type = graph_model::utility::function_collection<signature>;

  function_collection_type f;
  f.insert(T{0});
  f.insert(T{3});
  f.insert(T{1});
  f.insert(T{0});
  f.insert(T{0});

  auto i = f.begin<T>();
  BOOST_CHECK_EQUAL((*i++)(),0);
  BOOST_CHECK_EQUAL((*i++)(),1);
  BOOST_CHECK_EQUAL((*i++)(),3);
  BOOST_CHECK(i == f.end<T>());
}
#endif	// XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED

BOOST_AUTO_TEST_CASE(test_mixed_container)
{
  using signature = int();
  using function_collection_type = graph_model::utility::function_collection<signature>;

  function_collection_type f;

  f.insert(function{0});
  f.insert(function{3});
  f.insert(function{1});
  f.insert(function{0});
  f.insert(function{0});

  f.insert(copyable{10});
  f.insert(copyable{13});
  f.insert(copyable{11});
  f.insert(copyable{10});
  f.insert(copyable{10});

#if XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED
  f.insert(moveable{20});
  f.insert(moveable{23});
  f.insert(moveable{21});
  f.insert(moveable{20});
  f.insert(moveable{20});
#endif	// XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED

  {
    auto i = f.begin<function>();
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK_EQUAL((*i++)(),3);
    BOOST_CHECK_EQUAL((*i++)(),1);
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK(i == f.end<function>());
  }

  {
    auto i = f.begin<copyable>();
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK_EQUAL((*i++)(),13);
    BOOST_CHECK_EQUAL((*i++)(),11);
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK(i == f.end<copyable>());
  }

#if XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED
  {
    auto i = f.begin<moveable>();
    BOOST_CHECK_EQUAL((*i++)(),20);
    BOOST_CHECK_EQUAL((*i++)(),21);
    BOOST_CHECK_EQUAL((*i++)(),23);
    BOOST_CHECK(i == f.end<moveable>());
  }
#endif	// XXX_MOVEABLE_TYPES_ARE_NOT_SUPPORTED

  {
    // XXX - this may be fragile as the segment order is undefined
    auto i = f.begin();
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK_EQUAL((*i++)(),13);
    BOOST_CHECK_EQUAL((*i++)(),11);
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK_EQUAL((*i++)(),10);
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK_EQUAL((*i++)(),3);
    BOOST_CHECK_EQUAL((*i++)(),1);
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK_EQUAL((*i++)(),0);
    BOOST_CHECK(i == f.end());
  }

}
