/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

#define BOOST_TEST_MODULE graph_model_vertex_expression_function

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/vertex/expression_function.hpp>
#include <graph_model/vertex/expression_terminal.hpp>
#include <graph_model/vertex/expression.hpp>
#include <graph_model/vertex/lit.hpp>
#include <graph_model/concept/expression_archetype.hpp>
#include <graph_model/concept/expression_concept.hpp>
#include <graph_model/concept/expression_policy_archetype.hpp>

#include <boost/mpl/vector.hpp>

template < typename T >
struct terminal
{
  terminal () = default;
  explicit terminal (T t) : t_(t) {}
  auto operator () () const { return t_; }

private:
  T t_ = { 42 };
};

template < typename T >
auto transform_expression (graph_model::vertex::expression_terminal<terminal<T>> expr)
{ return graph_model::vertex::evaluate_terminal(expr); }

template < typename T >
auto transform_expression (graph_model::vertex::expression_terminal<terminal<T>&> expr)
{ return graph_model::vertex::evaluate_terminal(expr); }

template < typename T >
struct expression_policy
{
  auto eval () const { return T{42}; }
};

using test_types = boost::mpl::vector<int,long,float,double,long double>;

BOOST_AUTO_TEST_CASE(test_expression_archetype)
{
  using expression_type = graph_model::concept::expression_archetype<>;
  BOOST_CONCEPT_ASSERT((graph_model::concept::expression_concept<expression_type>));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_builtin_lvalue_terminal,T,test_types)
{
  using graph_model::vertex::make_expression_function;
  T t = { 42 };
  auto f = make_expression_function(t);
  BOOST_CHECK_EQUAL(f(),t);
  t = 13;
  BOOST_CHECK_EQUAL(f(),t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_builtin_rvalue_terminal,T,test_types)
{
  using graph_model::vertex::make_expression_function;
  auto f = make_expression_function(42);
  BOOST_CHECK_EQUAL(f(),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_expression_lvalue_terminal,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  T t = { 42 };
  auto expr = lit(t);
  auto f = make_expression_function(expr);
  BOOST_CHECK_EQUAL(f(),t);
  t = 13;
  BOOST_CHECK_EQUAL(f(),t);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_expression_rvalue_terminal,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  auto f = make_expression_function(lit(42));
  BOOST_CHECK_EQUAL(f(),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_expression_rvalue,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  auto f = make_expression_function(lit(42) / 6);
  BOOST_CHECK_EQUAL(f(),7);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_expression_expression,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  auto f = make_expression_function(lit(42) / 6);
  auto g = make_expression_function(f);
  BOOST_CHECK_EQUAL(g(),7);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_logical_comparison_operators,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  BOOST_CHECK(make_expression_function(lit(0) < T(1))());
  BOOST_CHECK(make_expression_function(lit(0) < terminal<T>(1))());
  BOOST_CHECK(!make_expression_function(lit(1) < T(0))());
  BOOST_CHECK(!make_expression_function(lit(1) < terminal<T>(0))());
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_logical_operators,T,test_types)
{
  using graph_model::vertex::lit;
  using graph_model::vertex::make_expression_function;
  BOOST_CHECK(make_expression_function(lit(1) && T(1))());
  BOOST_CHECK(make_expression_function(lit(1) && terminal<T>(1))());
  BOOST_CHECK(!make_expression_function(lit(1) && T(0))());
  BOOST_CHECK(!make_expression_function(lit(1) && terminal<T>(0))());
  BOOST_CHECK(make_expression_function(lit(1) || terminal<T>(0))());
  BOOST_CHECK(!make_expression_function(lit(0) || terminal<T>(0))());
  BOOST_CHECK(make_expression_function(!lit(0) || terminal<T>(0))());
}
