/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE graph_model_vertex_rvalue_reference_wrapper

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include <graph_model/vertex/rvalue_reference_wrapper.hpp>

// #include <debugging::object_tracker_base.hpp>

struct copyable_only
// : debugging::object_tracker_base<copyable_only>
{
  copyable_only () = default;
  copyable_only (copyable_only const&) = default;
  copyable_only& operator = (copyable_only const&) = default;

  operator int () const { return 42; }
};

struct moveable_only
// : debugging::object_tracker_base<moveable_only>
{
  moveable_only () = default;
  moveable_only (moveable_only&&) = default;
  moveable_only& operator = (moveable_only&&) = default;

  operator int () const { return 42; }
};

struct constructible_only
// : debugging::object_tracker_base<constructible_only>
{
  constructible_only () = default;
  constructible_only (constructible_only const&) = delete;
  constructible_only (constructible_only&&) = delete;
  constructible_only& operator = (constructible_only const&) = delete;
  constructible_only& operator = (constructible_only&&) = delete;

  operator int () const { return 42; }
};

template < typename T >
struct constructor
{
  static auto apply ()
  {
    using std::cref;
    using graph_model::vertex::cref;
    return cref(T(42));
  }
};

struct test
{
  template < typename T >
  static void cref (T&& t)
  {
    using std::cref;
    using graph_model::vertex::cref;

    auto c_wrapper = cref(std::forward<T>(t));
    test_wrapper(c_wrapper);
  }

  template < typename T >
  static void ref (T&& t)
  {
    using std::ref;
    using graph_model::vertex::ref;

    auto wrapper = ref(std::forward<T>(t));
    test_wrapper(wrapper);
  }

  template < typename T >
  static void apply (T&& t)
  {
    using value_type = std::decay_t<T>;
    using wrapper_type = graph_model::vertex::rvalue_reference_wrapper<value_type>;
    wrapper_type wrapper(std::forward<T>(t));
    test_wrapper(wrapper);
  }

private:
  template < typename Wrapper >
  static void test_wrapper (Wrapper& wrapper)
  {
    auto& x = wrapper.get();
    BOOST_CHECK_EQUAL(wrapper.get(),42);
    Wrapper other (wrapper);
    BOOST_CHECK_EQUAL(wrapper.get(),42);
    BOOST_CHECK_EQUAL(other.get(),42);
    other = wrapper;
    BOOST_CHECK_EQUAL(wrapper.get(),42);
    BOOST_CHECK_EQUAL(other.get(),42);
    Wrapper other2 (std::move(other));
    BOOST_CHECK_EQUAL(other2.get(),42);
    other2 = std::move(wrapper);
    BOOST_CHECK_EQUAL(other2.get(),42);
  }
};

#include <boost/mpl/vector.hpp>

using test_types = boost::mpl::vector<int,long,long long,float,double,long double>;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_rvalue_reference_wrapper,T,test_types)
{
  int t = { 42 };
  copyable_only copyable;
  moveable_only moveable;
  copyable_only const c_copyable = {};
  moveable_only const c_moveable = {};
  constructible_only constructible;
  test::apply(42);
  test::apply(T{42});
  test::apply(t);
  test::apply(copyable);
  test::apply(copyable_only());
  test::apply(c_copyable);
  test::apply(std::move(moveable));
  test::apply(moveable_only());
#if XXX_NO_SUITABLE_CONSTRUCTORS
  test::apply(std::move(c_moveable));
  test::apply(constructible);
  test::apply(std::move(constructible));
  test::apply(constructible_only());
#endif	// XXX_NO_SUITABLE_CONSTRUCTORS
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_cref,T,test_types)
{
  T t = { 42 };
  T& ref = t;
  T const& cref = t;
  copyable_only x;
  moveable_only y;
  test::cref(t);
  test::cref(T(42));
  test::cref(42);
  test::cref(ref);
  test::cref(cref);
  test::cref(x);
  test::cref(copyable_only());
  test::cref(y);
  test::cref(moveable_only());
  test::cref(constructor<T>::apply());
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_ref,T,test_types)
{
  T t = { 42 };
  T& ref = t;
  T const& cref = t;
  copyable_only x;
  moveable_only y;
  test::ref(t);
  test::ref(T(42));
  test::ref(42);
  test::ref(ref);
  test::ref(cref);
  test::ref(x);
  test::ref(copyable_only());
  test::ref(y);
  test::ref(moveable_only());
}

