/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#define BOOST_TEST_MODULE sample_space_vector_simplex

#ifdef INCLUDE_UNIT_TEST_HEADERS
#include <boost/test/included/unit_test.hpp>
#else
#include <boost/test/unit_test.hpp>
#endif

template < typename ... T > class XXX;

#include "sample_space/vector/simplex.hpp"
#include "sample_space/math/norm_1.hpp"
#include <boost/mpl/vector.hpp>

typedef boost::mpl::vector<char,int,long,unsigned,unsigned long,float,double,long double> test_types;
typedef boost::mpl::vector<signed char,int,long,float,double,long double> signed_test_types;
typedef boost::mpl::vector<char,int,long,unsigned,unsigned long> integer_test_types;
typedef boost::mpl::vector<float,double,long double> floating_point_test_types;

BOOST_AUTO_TEST_CASE_TEMPLATE(test_typedef,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_default_constructor,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;
  
  simplex_type simplex;
  BOOST_CHECK(simplex.empty());
  BOOST_CHECK_EQUAL(simplex.size(),0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type simplex (42);
  BOOST_CHECK(!simplex.empty());
  BOOST_CHECK_EQUAL(simplex.size(),42);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor_2,T,floating_point_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type simplex (13,42);
  BOOST_CHECK(!simplex.empty());
  BOOST_CHECK_EQUAL(simplex.size(),13);
  BOOST_CHECK_EQUAL(*simplex.begin(),42);
  BOOST_CHECK(!is_possible(simplex));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_constructor_3,T,floating_point_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type simplex = { 1, 2, 3 };
  BOOST_CHECK(!simplex.empty());
  BOOST_CHECK_EQUAL(simplex.size(),3);
  BOOST_CHECK_EQUAL(*simplex.begin(),1);
  BOOST_CHECK(!is_possible(simplex));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_constructor,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v0;
  simplex_type v(v0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_constructor,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v0;
  simplex_type v(std::move(v0));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_copy_assignment,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v0;
  simplex_type v1;
  v1 = v0;
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_move_assignment,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v0;
  simplex_type v1;
  v1 = std::move(v0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_at,T,floating_point_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type simplex (3);
  simplex.at(2) = 0.25;
  simplex.at(1) = 0.50;
  simplex.at(0) = 0.25;
  BOOST_CHECK_EQUAL(simplex.at(0),0.25);
  BOOST_CHECK_EQUAL(simplex.at(1),0.50);
  BOOST_CHECK_EQUAL(simplex.at(2),0.25);
  BOOST_CHECK(is_possible(simplex));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_equal,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type simplex (13,42);
  BOOST_CHECK_EQUAL(simplex,simplex);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_less_than,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v1 (13,42);
  simplex_type v2 (14,42);
  BOOST_CHECK(v1 < v2);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_add,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v1 (13,1);
  simplex_type v2 (13,2);
  simplex_type expected (13,3);
  BOOST_CHECK_EQUAL(v1 + v2, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_subtract,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v1 (13,1);
  simplex_type v2 (13,2);
  simplex_type expected (13,1);
  BOOST_CHECK_EQUAL(v2 - v1, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_multiply,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v (13,42);
  T x = 2;
  simplex_type expected (13,42 * 2);
  BOOST_CHECK_EQUAL(v * x, expected);
  BOOST_CHECK_EQUAL(x * v, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_divide,T,test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v (13,42);
  T x = 7;
  simplex_type expected (13,42 / 7);
  BOOST_CHECK_EQUAL(v / x, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_modulo,T,integer_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  simplex_type v (13,42);
  T x = 10;
  simplex_type expected (13,42 % 10);
  BOOST_CHECK_EQUAL(v % x, expected);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_norm_1,T,signed_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  using std::abs;
  using sample_space::math::norm_1;
  
  simplex_type simplex (3);
  BOOST_CHECK_EQUAL(simplex.size(),3);
  simplex.at(0) = 0;
  simplex.at(1) = 1;
  simplex.at(2) = -3;
  BOOST_CHECK_EQUAL(norm_1(simplex),abs(T(0))+abs(T(1))+abs(T(-3)));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_norm_2,T,signed_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  using std::sqrt;

  simplex_type simplex (3);
  simplex.at(0) = 0;
  simplex.at(1) = 1;
  simplex.at(2) = -3;
  BOOST_CHECK_EQUAL(norm_2(simplex),T(sqrt(T(0)*T(0)+T(1)*T(1)+T(-3)*T(-3))));
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_zero,T,signed_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  using std::sqrt;

  simplex_type simplex (3);
  simplex.at(0) = 0;
  simplex.at(1) = 1;
  simplex.at(2) = -3;

  auto z = zero(simplex);
  BOOST_CHECK_EQUAL(z.at(0),0);
  BOOST_CHECK_EQUAL(z.at(1),0);
  BOOST_CHECK_EQUAL(z.at(2),0);
}

BOOST_AUTO_TEST_CASE_TEMPLATE(test_one,T,signed_test_types)
{
  using simplex_type = sample_space::vector::simplex<T>;

  using std::sqrt;

  simplex_type simplex (3);
  simplex.at(0) = 0;
  simplex.at(1) = 1;
  simplex.at(2) = -3;

  auto o = one(simplex);
  BOOST_CHECK_EQUAL(o.at(0),1);
  BOOST_CHECK_EQUAL(o.at(1),1);
  BOOST_CHECK_EQUAL(o.at(2),1);
}
